/************************************************************************
 *   IRC - Internet Relay Chat, src/channel.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "channel.h"
#include "client.h"
#include "hash.h"
#include "ircd.h"
#include "list.h"
#include "numeric.h"
#include "send.h"
#include "struct.h"
#include "flud.h"
#include "memory.h"
#include "irc_string.h"
#include "common.h"

#ifdef NEED_SPLITCODE
#if 0
int server_was_split = YES;
int got_server_pong;
time_t server_split_time;
#endif
#endif

time_t next_temp_invite_due = -1;

struct Channel *channel = NullChn;

/* 
 * return the length (>=0) of a chain of links.
 */
static int
list_length(Link * lp)
{
  int count = 0;

  for (; lp; lp = lp->next)
    count++;
  return count;
}

/*
 *  Fixes a string so that the first white space found becomes an end of
 * string marker (`\0`).  returns the 'fixed' string or "*" if the string
 * was NULL length or a NULL pointer.
 */
static char *
check_string(char *s)
{
  static char star[2] = "*";
  char *str = s;

  if (BadPtr(s))
    return star;

  for (; *s; ++s)
    {
      if (*s == ' ')
	{
	  *s = '\0';
	  break;
	}
    }
  return str;
}

/*
 * create a string of form "foo!bar@fubar" given foo, bar and fubar
 * as the parameters.  If NULL, they become "*".
 */
static char *
make_nick_user_host(const char *nick, const char *name, const char *host)
{
  static char namebuf[NICKLEN + USERLEN + HOSTLEN + 6];
  int n;
  char *s;
  const char *p;

  s = namebuf;

  for (p = nick, n = NICKLEN; *p && n--;)
    *s++ = *p++;
  /* This hackery prevents the problem when a ban of *nick or whatever
   * is placed, and the last character is clipped, making a non-matching
   * mask
   *
   * This changes the last character into a * if it is truncated
   */
  if (*p)
    s[-1] = '*';
  *s++ = '!';
  for (p = name, n = USERLEN; *p && n--;)
    *s++ = *p++;
  if (*p)
    s[-1] = '*';
  *s++ = '@';
  for (p = host, n = HOSTLEN; *p && n--;)
    *s++ = *p++;
  if (*p)
    s[-1] = '*';
  *s = '\0';
  return namebuf;
}

/*
 * del_matching_exception - delete an exception matching this user
 *
 * The idea is, if a +e client gets kicked for any reason
 * remove the matching exception for this client.
 * This will immediately stop channel "chatter" with scripts
 * that kick on matching ban. It will also stop apparent "desyncs."
 * It's not the long term answer, but it will tide us over.
 *
 * modified from orabidoo - Dianora
 */
static void
del_matching_exception(struct Client *cptr, struct Channel *chptr)
{
  struct SLink *exception;
  struct channel_mask *mask;
  int type = cmodes[cmode_index[(unsigned char) 'e']].value;

  if (!IsPerson(cptr))
    return;

  mask = build_mask(cptr);

  while ((exception = find_channel_mask(mask, chptr, type, 1, 1)))
    {
      sendto_channel_butserv(chptr, &me, ":%s MODE %s -e %s!%s@%s",
			     me.name, chptr->chname,
			     exception->value.cmptr->nick,
			     exception->value.cmptr->user,
			     exception->value.cmptr->host);
      sendto_match_servs(chptr, &me, ":%s MODE %s -e %s!%s@%s",
			 me.name, chptr->chname,
			 exception->value.cmptr->nick,
			 exception->value.cmptr->user,
			 exception->value.cmptr->host);
      delete_channel_mask(chptr, exception, type);
    }
}

int
is_invex(struct Client *cptr, struct Channel *chptr)
{
  return (match_channel_mask(build_mask(cptr), chptr, 'I') != NULL);
}

/*
 * adds a user to a channel by adding another link to the channels member
 * chain.
 */
void
add_user_to_channel(struct Channel *chptr, struct Client *who, int flags)
{
  Link *ptr;

  if (who->user)
    {
      ptr = make_link();
      ptr->flags = flags;
      ptr->value.cptr = who;
      ptr->next = chptr->members;
      chptr->members = ptr;

      chptr->users++;

      ptr = make_link();
      ptr->value.chptr = chptr;
      ptr->next = who->user->channel;
      who->user->channel = ptr;
      who->user->joined++;
    }
}

void
remove_user_from_channel(struct Client *sptr, struct Channel *chptr,
			 int was_kicked)
{
  Link **curr;
  Link *tmp;

  for (curr = &chptr->members; (tmp = *curr); curr = &tmp->next)
    if (tmp->value.cptr == sptr)
      {
	/* User was kicked, but had an exception.
	 * so, to reduce chatter I'll remove any
	 * matching exception now.
	 */
	if (was_kicked && (tmp->flags & CUFLAG_EXCEPTION))
	  {
	    del_matching_exception(sptr, chptr);
	  }
	*curr = tmp->next;
	free_link(tmp);
	break;
      }
  for (curr = &sptr->user->channel; (tmp = *curr); curr = &tmp->next)
    if (tmp->value.chptr == chptr)
      {
	*curr = tmp->next;
	free_link(tmp);
	break;
      }
  for (curr = &chptr->tmemberlist; (tmp = *curr); curr = &tmp->next)
    if (tmp->value.tiptr->cptr == sptr)
      {
	*curr = tmp->next;
	MyFree(tmp->value.tiptr);
	free_link(tmp);
      }
  sptr->user->joined--;

  sub1_from_channel(chptr);
}

int
is_chan_op(struct Client *cptr, struct Channel *chptr)
{
  Link *lp;

  if (chptr)
    if ((lp = find_user_link(chptr->members, cptr)))
      return (lp->flags & CUFLAG_CHANOP);

  return 0;
}

int
has_voice(struct Client *cptr, struct Channel *chptr)
{
  Link *lp;

  if (chptr)
    if ((lp = find_user_link(chptr->members, cptr)))
      return (lp->flags & CUFLAG_VOICE);

  return 0;
}

int
can_send(struct Client *cptr, struct Channel *chptr)
{
  Link *lp;

  if (HasUmode(cptr, UMODE_GOD))
    return 0;
#ifdef JUPE_CHANNEL
  if (MyClient(cptr) && (chptr->mode_flags & CMODE_JUPED))
    {
      return (CMODE_JUPED);
    }
#endif

  lp = find_user_link(chptr->members, cptr);

  if (chptr->mode_flags & CMODE_MODERATED &&
      (!lp || !(lp->flags & (CUFLAG_CHANOP | CUFLAG_VOICE))))
    return (CMODE_MODERATED);

  if (chptr->mode_flags & CMODE_NOPRIVMSGS && !lp)
    return (CMODE_NOPRIVMSGS);

  if (match_channel_mask(build_mask(cptr), chptr, 'q') &&
      (!lp || !(lp->flags & (CUFLAG_CHANOP | CUFLAG_VOICE))))
    return (CMODE_MODERATED);

  return 0;
}

/* stolen from Undernet's ircd  -orabidoo
 *
 */

char *
pretty_mask(char *mask)
{
  register char *cp = mask;
  register char *user;
  register char *host;

  if ((user = strchr(cp, '!')))
    *user++ = '\0';
  if ((host = strrchr(user ? user : cp, '@')))
    {
      *host++ = '\0';
      if (!user)
	return make_nick_user_host("*", check_string(cp), check_string(host));
    }
  else if (!user && strchr(cp, '.'))
    return make_nick_user_host("*", "*", check_string(cp));
  return make_nick_user_host(check_string(cp), check_string(user),
			     check_string(host));
}

int
can_join(struct Client *sptr, struct Channel *chptr, char *key, int *flags)
{
  struct channel_mask *mask;
  int limit;

  if (HasUmode(sptr, UMODE_GOD))
    return 0;

#ifdef JUPE_CHANNEL
  if (chptr->mode_flags & CMODE_JUPED)
    {
      sendto_ops_flag(UMODE_SPY, get_str(STR_TRY_JOIN_JUPE),	/* "User %s (%s@%s) is attempting to join juped channel %s" */
		      sptr->name, sptr->username, sptr->host, chptr->chname);
      return (ERR_UNAVAILRESOURCE);
    }
#endif

  mask = build_mask(sptr);
  if (match_channel_mask(mask, chptr, 'b'))
    {
      if (match_channel_mask(mask, chptr, 'e'))
	{
	  change_chan_flag(chptr, sptr, CUFLAG_EXCEPTION, MODE_ADD);
	  return 0;
	}
      else
	return ERR_BANNEDFROMCHAN;
    }

  if ((chptr->mode_flags & CMODE_INVITEONLY) && !is_invex(sptr, chptr))
    {
      struct SLink *lp;
      for (lp = sptr->user->invited; lp; lp = lp->next)
	if (lp->value.chptr == chptr)
	  break;
      if (!lp)
	return (ERR_INVITEONLYCHAN);
    }

  limit = get_mode_intval(chptr, 'l');
  if (limit && chptr->users >= limit)
    return (ERR_CHANNELISFULL);

  return 0;
}

/*
 * check_channel_name - check channel name for invalid characters
 * return true (1) if name ok, false (0) otherwise
 */
int
check_channel_name(const char *name)
{
  if (!name)
    return 0;

  for (; *name; ++name)
    {
      if (!IsChanChar(*name))
	return 0;
    }
  return 1;
}

/*
**  Get Channel block for chname (and allocate a new channel
**  block, if it didn't exist before).
*/
struct Channel *
get_channel(struct Client *cptr, char *chname, int flag)
{
  struct Channel *chptr;
  int len;

  if (BadPtr(chname))
    return NULL;

  len = strlen(chname);
  if (MyClient(cptr) && len > CHANNELLEN)
    {
      len = CHANNELLEN;
      *(chname + CHANNELLEN) = '\0';
    }
  if ((chptr = hash_find_channel(chname, NULL)))
    return (chptr);

  /*
   * If a channel is created during a split make sure its marked
   * as created locally 
   * Also make sure a created channel has =some= timestamp
   * even if it get over-ruled later on. Lets quash the possibility
   * an ircd coder accidentally blasting TS on channels. (grrrrr -db)
   *
   * Actually, it might be fun to make the TS some impossibly huge value (-db)
   */

  if (flag == CREATE)
    {
      chptr = (struct Channel *) MyMalloc(sizeof(struct Channel) + len + 1);
      memset(chptr, 0, sizeof(struct Channel));
      /*
       * NOTE: strcpy ok here, we have allocated strlen + 1
       */
      strcpy(chptr->chname, chname);
      if (channel)
	channel->prevch = chptr;
      chptr->prevch = NULL;
      chptr->nextch = channel;
      channel = chptr;
      chptr->channelts = CurrentTime;	/* doesn't hurt to set it here */
      chptr->remotets = -1;	/* We don't have a burst coming on this channel (yet) */
      add_to_channel_hash_table(chname, chptr);
      Count.chan++;
    }
  return chptr;
}

void
add_invite(struct Client *cptr, struct Channel *chptr, time_t until)
{
  Link *inv, **tmp;

  del_invite(cptr, chptr);
  /*
   * delete last link in chain if the list is max length
   */
  if (list_length(cptr->user->invited) >= MAXCHANNELSPERUSER)
    {
      del_invite(cptr, cptr->user->invited->value.chptr);
    }
  /*
   * add client to channel invite list
   */
  inv = make_link();
  inv->value.tiptr = (struct TInvite *) MyMalloc(sizeof(struct TInvite));
  inv->value.tiptr->cptr = cptr;
  inv->value.tiptr->until = until;
  inv->next = chptr->invites;
  chptr->invites = inv;
  /*
   * add channel to the end of the client invite list
   */
  for (tmp = &(cptr->user->invited); *tmp; tmp = &((*tmp)->next))
    ;
  inv = make_link();
  inv->value.chptr = chptr;
  inv->next = NULL;
  (*tmp) = inv;
}

/*
 * Delete Invite block from channel invite list and client invite list
 */
void
del_invite(struct Client *cptr, struct Channel *chptr)
{
  Link **inv, *tmp;

  for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next)
    if (tmp->value.tiptr->cptr == cptr)
      {
	*inv = tmp->next;
	MyFree(tmp->value.tiptr);
	free_link(tmp);
	break;
      }

  for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next)
    if (tmp->value.chptr == chptr)
      {
	*inv = tmp->next;
	free_link(tmp);
	break;
      }
}

/*
**  Subtract one user from channel (and free channel
**  block, if channel became empty).
*/
void
sub1_from_channel(struct Channel *chptr)
{
  Link *tmp;

  if (--chptr->users <= 0)
    {
      chptr->users = 0;		/* if chptr->users < 0, make sure it sticks at 0
				 * It should never happen but...
				 *
				 * Actually it can now happen, during CLABORT
				 *  -- asuffield
				 */

      if (chptr->mode_flags & CMODE_PERM)
	/* This is a permanent channel. Do not delete it. */
	return;

#ifdef JUPE_CHANNEL
      if (chptr->mode_flags & CMODE_JUPED)
	{
	  while ((tmp = chptr->invites))
	    del_invite(tmp->value.tiptr->cptr, chptr);

#ifdef FLUD
	  free_fluders(NULL, chptr);
#endif
	}
      else
#endif
	{
	  /*
	   * Now, find all invite links from channel structure
	   */
	  while ((tmp = chptr->invites))
	    del_invite(tmp->value.tiptr->cptr, chptr);

	  /* free all bans/exceptions/denies */
	  free_channel_modes(chptr);

	  if (chptr->prevch)
	    chptr->prevch->nextch = chptr->nextch;
	  else
	    channel = chptr->nextch;
	  if (chptr->nextch)
	    chptr->nextch->prevch = chptr->prevch;

#ifdef FLUD
	  free_fluders(NULL, chptr);
#endif
	  del_from_channel_hash_table(chptr->chname, chptr);
	  MyFree((char *) chptr);
	  Count.chan--;
	}
    }
}

int
count_channels(struct Client *sptr)
{
  struct Channel *chptr;
  int count = 0;

  for (chptr = channel; chptr; chptr = chptr->nextch)
    count++;
  return (count);
}

struct Channel *
get_ban_channel_forward(struct Client *cptr, struct Channel *chptr)
{
  struct SLink *cmlink;

  if (!IsPerson(cptr))
    return NULL;

  cmlink = match_channel_mask(build_mask(cptr), chptr, 'b');
  return get_channel(cptr, cmlink->value.cmptr->channel, !CREATE);
}
